import pika
import functools
from threading import Thread

import logging
LOGGER = logging.getLogger(__name__)

class MessageReceiver:

    _exchange_name = None
    _exchange_type = None
    _rabbit_mq_ip = None
    _rabbit_mq_port = None
    _auto_ack = None
    _queue_name = None

    _connection: pika.SelectConnection = None
    _message_received_callback = None

    def __init__(self, exchange_name='beacon', exchange_type='fanout', rabbit_mq_ip='143.205.173.36', rabbit_mq_port=30302, queue_name='', auto_ack=True):
        self._exchange_name = exchange_name
        self._exchange_type = exchange_type
        self._rabbit_mq_ip = rabbit_mq_ip
        self._rabbit_mq_port = rabbit_mq_port
        self._auto_ack = auto_ack
        self._queue_name = queue_name

    #region Connection establishment

    def _connect(self) -> pika.SelectConnection:
        connection = pika.SelectConnection(parameters=pika.ConnectionParameters(self._rabbit_mq_ip, self._rabbit_mq_port), 
            on_open_callback=self._connection_opened_callback, on_open_error_callback=self._connection_opened_error)
        return connection

    def _connection_opened_callback(self, connection: pika.SelectConnection):
        # Create channel
        connection.channel(on_open_callback=self._channel_created_callback)

    def _connection_opened_error(self, connection, err):
        LOGGER.error(f"RabbitMQ connection could not be established: {str(err)}")

    def _channel_created_callback(self, channel: pika.channel.Channel):
        # Create exchange
        cb = functools.partial(self._exchange_created_callback, userdata=channel)
        channel.exchange_declare(exchange=self._exchange_name, exchange_type=self._exchange_type, callback=cb)

    def _exchange_created_callback(self, answer_message, userdata):
        # Create queue
        channel = userdata
        cb = functools.partial(self._queue_created_callback, userdata=channel)
        channel.queue_declare(queue=self._queue_name, exclusive=False, callback=cb)

    def _queue_created_callback(self, answer_message: pika.frame.Method, userdata):
        queue_name = answer_message.method.queue
        channel = userdata
        
        # Bind queue to exchange
        channel.queue_bind(exchange=self._exchange_name, queue=queue_name)
        LOGGER.info(f"RabbitMQ connection to exchange '{self._exchange_name}' established")

        # Consume messages
        channel.basic_consume(queue=queue_name, auto_ack=self._auto_ack, on_message_callback=self._message_received_callback)

    #endregion Connection establishment

    def start(self, message_received_callback):
        '''Connects to RabbitMQ and starts listening for messages'''
        self._message_received_callback = message_received_callback

        self._connection = self._connect()
        Thread(target=self._connection.ioloop.start).start()

    def stop(self):
        '''Stops listening for messages and closes the connection'''
        self._connection.ioloop.stop()
        self._connection.close()
        LOGGER.info(f"RabbitMQ connection closed")